@@ -10,6 +10,7 @@ from typing import (
     TYPE_CHECKING,
     Any,
     Dict,
+    Iterable,
     List,
     Mapping,
     Optional,
@@ -42,6 +43,7 @@ from django.dispatch.dispatcher import receiver
 from django.forms.widgets import Widget
 from django.utils.functional import cached_property
 from django.utils.text import slugify
+from flexible_forms.signals import pre_form_class_prepare, post_form_class_prepare
 from simpleeval import FunctionNotDefined, NameNotDefined
 
 from flexible_forms.fields import FIELD_TYPES, FieldType
@@ -464,7 +466,7 @@ class BaseForm(FlexibleBaseModel):
         abstract = True
 
     def __str__(self) -> str:
-        return self.label or f"Untitled {self.__class__.__name__}"
+        return self.label or f"Untitled {cast(str, self._meta.verbose_name).title()}"
 
     def save(self, *args: Any, **kwargs: Any) -> None:
         """Save the model.
@@ -546,29 +548,28 @@ class BaseForm(FlexibleBaseModel):
 
     def as_django_form(
         self,
-        data: Optional[Mapping[str, Any]] = None,
-        files: Optional[Mapping[str, Any]] = None,
+        data: Optional[Dict[str, Any]] = None,
+        files: Optional[Dict[str, Any]] = None,
         instance: Optional["BaseRecord"] = None,
         initial: Optional[Mapping[str, Any]] = None,
+        exclude: Iterable[str] = (),
         **kwargs: Any,
     ) -> "BaseRecordForm":
         """Return the form represented as a Django form instance.
 
         Args:
-            data: Data to be passed to the
-                Django form constructor.
-            files: Data to be passed to the
-                Django form constructor.
-            instance: The Record instance to be passed to
-                the Django form constructor.
-            initial: Initial values for the
-                form fields.
+            data: Data to be passed to the Django form constructor.
+            files: Data to be passed to the Django form constructor.
+            instance: The Record instance to be passed to the Django form
+                constructor.
+            initial: Initial values for the form fields.
+            exclude: Field names to exclude from the form.
             **kwargs: Passed to the Django ModelForm constructor.
 
         Returns:
             BaseRecordForm: A configured BaseRecordForm (a Django ModelForm instance).
         """
-        all_fields = self.fields.all()
+        all_fields = tuple(f for f in self.fields.all() if f.name not in exclude)
 
         # Build a dict containing all field values. This combines all of the
         # form data into a single structure that will be used when evaluating
@@ -592,6 +593,18 @@ class BaseForm(FlexibleBaseModel):
         # Import the form class inline to prevent a circular import.
         from flexible_forms.forms import BaseRecordForm
 
+        # Send a signal before preparing the form class to give receivers an
+        # opportunity to modify the form structure.
+        pre_form_class_prepare.send(
+            sender=self.__class__,
+            data=data,
+            files=files,
+            instance=instance,
+            initial=initial,
+            exclude=exclude,
+            form_fields=form_fields,
+        )
+
         form_name = f"{self.name.title().replace('_', '')}Form"
         form_class: Type[BaseRecordForm] = type(
             form_name,
@@ -601,12 +614,28 @@ class BaseForm(FlexibleBaseModel):
                 "Meta": type(
                     "Meta",
                     (BaseRecordForm.Meta,),
-                    {"model": RecordModel},
+                    {
+                        "model": RecordModel,
+                        "exclude": exclude,
+                    },
                 ),
                 **form_fields,
             },
         )
 
+        # Send a signal before preparing the form class to give receivers an
+        # opportunity to modify the form class after creation.
+        post_form_class_prepare.send(
+            sender=self.__class__,
+            data=data,
+            files=files,
+            instance=instance,
+            initial=initial,
+            exclude=exclude,
+            form_fields=form_fields,
+            form_class=form_class,
+        )
+
         initial = {"form": self, **(initial or {})}
 
         # Create a form instance from the form class and the passed parameters.
@@ -669,6 +698,13 @@ class BaseField(FlexibleBaseModel):
         help_text="The form widget to use when displaying the field.",
     )
 
+    field_type_options = JSONField(
+        blank=True,
+        default=dict,
+        help_text="Custom configuration for field types (only used for custom fields).",
+        encoder=DjangoJSONEncoder,
+    )
+
     form_field_options = JSONField(
         blank=True,
         default=dict,
@@ -704,7 +740,7 @@ class BaseField(FlexibleBaseModel):
         form_field_related_name = "fields"
 
     def __str__(self) -> str:
-        return self.label or "New Field"
+        return self.label or f"New {cast(str, self._meta.verbose_name).title()}"
 
     def save(self, *args: Any, **kwargs: Any) -> None:
         """Save the model.
@@ -720,18 +756,32 @@ class BaseField(FlexibleBaseModel):
 
         super().save(*args, **kwargs)
 
-    def as_field_type(self) -> Type[FieldType]:
+    def as_field_type(
+        self,
+        record: Optional["BaseRecord"] = None,
+        field_values: Optional[Dict[str, Any]] = None,
+    ) -> FieldType:
         """Return the flexible FieldType for the field.
 
+        Args:
+            record: The record instance bound to the form.
+            field_values: The current values of the fields in the form.
+
         Returns:
             FieldType: The FieldType class for the field.
         """
-        return FIELD_TYPES[self.field_type]
+        return FIELD_TYPES[self.field_type](
+            field=self,
+            record=record,
+            field_values=field_values,
+            modifiers=tuple((m.attribute, m.expression) for m in self.modifiers.all()),
+            **self.field_type_options,
+        )
 
     def as_form_field(
         self,
-        field_values: Dict[str, Any],
         record: Optional["BaseRecord"] = None,
+        field_values: Optional[Dict[str, Any]] = None,
     ) -> forms.Field:
         """Return a Django form Field instance.
 
@@ -743,37 +793,30 @@ class BaseField(FlexibleBaseModel):
         Returns:
             forms.Field: The configured Django form Field instance.
         """
-        return self.as_field_type().as_form_field(
-            field=self,
-            field_values=field_values,
-            record=record,
-            modifiers=tuple((m.attribute, m.expression) for m in self.modifiers.all()),
-            # Django form field arguments.
-            required=self.required,
-            label=self.label,
-            label_suffix=self.label_suffix,
-            initial=self.initial,
-            help_text=self.help_text,
-            widget=self.as_form_widget(record=record),
-            **self.form_field_options,
-        )
+        return self.as_field_type(
+            record=record, field_values=field_values
+        ).as_form_field()
 
-    def as_form_widget(self, record: Optional["BaseRecord"] = None) -> Widget:
+    def as_form_widget(
+        self,
+        record: Optional["BaseRecord"] = None,
+        field_values: Optional[Dict[str, Any]] = None,
+    ) -> Widget:
         """Return a Django form Widget instance.
 
         Uses the form_widget_options and the given field_values to customize
         the widget.
 
         Args:
-            record: The record instance that the form is bound to, if
-                available.
+            record: The record instance bound to the form.
+            field_values: The current values of the fields in the form.
 
         Returns:
             Widget: The configured Django form Widget instance.
         """
-        return self.as_field_type().as_form_widget(
-            field=self, record=record, **self.form_widget_options
-        )
+        return self.as_field_type(
+            record=record, field_values=field_values
+        ).as_form_widget()
 
     def as_model_field(self) -> models.Field:
         """Return a Django model Field definition.
@@ -781,12 +824,7 @@ class BaseField(FlexibleBaseModel):
         Returns:
             models.Field: The configured Django model Field instance.
         """
-        return self.as_field_type().as_model_field(
-            null=not self.required,
-            blank=not self.required,
-            default=self.initial,
-            help_text=self.help_text,
-        )
+        return self.as_field_type().as_model_field()
 
 
 class BaseFieldModifier(FlexibleBaseModel):
@@ -839,7 +877,7 @@ class BaseFieldModifier(FlexibleBaseModel):
         field_field_related_name = "modifiers"
 
     def __str__(self) -> str:
-        return f"{self.attribute} = {self.expression}"
+        return f"{cast(str, self._meta.verbose_name).title()} ({self.attribute} = {self.expression})"
 
     def clean(self) -> None:
         """Ensure that the expression is valid for the form.
@@ -972,7 +1010,12 @@ class BaseFieldset(FlexibleBaseModel):
         form_field_related_name = "fieldsets"
 
     def __str__(self) -> str:
-        return f"Fieldset {self.name} ({self.pk})"
+        return " ".join(
+            filter(
+                bool,
+                [cast(str, self._meta.verbose_name).title(), self.name, f"({self.pk})"],
+            )
+        )
 
 
 class BaseFieldsetItem(FlexibleBaseModel):
@@ -1018,7 +1061,7 @@ class BaseFieldsetItem(FlexibleBaseModel):
         fieldset_field_related_name = "items"
 
     def __str__(self) -> str:
-        return f"Fieldset item {self.pk} for field {self.field_id}"
+        return f"{cast(str, self._meta.verbose_name).title()} {self.pk} for field {self.field_id}"
 
 
 class RecordManager(models.Manager):
@@ -1107,16 +1150,14 @@ class BaseRecord(FlexibleBaseModel):
         record_type = (
             self.form.label
             if self.__dict__.get("form")
-            or self._meta.get_field("form").get_cached_value(self)
-            else "Record"
+            or self._meta.get_field("form").get_cached_value(self, default=None)
+            else cast(str, self._meta.verbose_name).title()
         )
 
         return f"New {record_type}" if not self.pk else f"{record_type} {self.pk}"
 
-    def __init__(
-        self, *args: Any, form: Optional[BaseForm] = None, **kwargs: Any
-    ) -> None:
-        super().__init__(*args, form=form, **kwargs)
+    def __init__(self, *args: Any, **kwargs: Any) -> None:
+        super().__init__(*args, **kwargs)
         self._staged_changes = {}
         self._initialized = True
 
@@ -1147,7 +1188,7 @@ class BaseRecord(FlexibleBaseModel):
             return {}
         return {f.name: f for f in self.form.fields.all()}
 
-    @cached_property
+    @property
     def _data(self) -> Mapping[str, Any]:
         """Return a dict of Record attributes and their values.
 
@@ -1320,7 +1361,7 @@ class BaseRecordAttribute(FlexibleBaseModel):
         record_field_related_name = "attributes"
 
     def __str__(self) -> str:
-        return f"RecordAttribute {self.pk} (record_id={self.record_id}, field_id={self.field_id})"
+        return f"{cast(str, self._meta.verbose_name).title()} {self.pk}"
 
     @classmethod
     def get_value_field_name(cls, field_type: str) -> str:
@@ -1390,7 +1431,7 @@ class BaseRecordAttribute(FlexibleBaseModel):
 # field.
 #
 for field_type_name, field_type in sorted(FIELD_TYPES.items(), key=lambda f: f[0]):
-    BaseRecordAttribute.add_to_class(  # type: ignore
+    BaseRecordAttribute.add_to_class(
         BaseRecordAttribute.get_value_field_name(field_type_name),
         field_type.as_model_field(blank=True, null=True, default=None),
     )
