@@ -11,17 +11,14 @@ from typing import (
     Sequence,
     Tuple,
     Union,
+    cast,
 )
 
 from django.contrib.admin.widgets import AutocompleteMixin
 from django.forms.widgets import Select
-from typing_extensions import TypedDict
 
 from flexible_forms.utils import stable_json
 
-AutocompleteResult = TypedDict("AutocompleteResult", {"id": str, "text": str})
-AutocompleteValue = Union[AutocompleteResult, Sequence[AutocompleteResult]]
-
 
 class AutocompleteSelect(Select):
     """A select widget where options are sourced from a remote URL."""
@@ -46,8 +43,8 @@ class AutocompleteSelect(Select):
         self.allow_freetext = allow_freetext
         self.placeholder = placeholder or ""
 
-    def optgroups(
-        self, name: str, value: List[str], *args: Any, **kwargs: Any
+    def optgroups(  # type: ignore
+        self, name: str, value: str, attrs: Optional[Dict[Any, Any]] = None
     ) -> List[Tuple[None, List[Dict[str, Any]], int]]:
         """Build a list of optgroups for populating the select.
 
@@ -62,8 +59,7 @@ class AutocompleteSelect(Select):
         Args:
             name: The name of the widget in the form.
             value: The value of the field.
-            args: Unused.
-            kwargs: Unused.
+            attrs: Unused.
 
         Returns:
             List[Tuple[Optional[str], List[Tuple[Any, Any]], int]]: Optgroups
@@ -73,21 +69,31 @@ class AutocompleteSelect(Select):
         groups = [default]
         has_selected = False
 
+        # Parse the given value as a JSON string, and make it a list if it
+        # isn't already so that we can iterate over it when we enumerate
+        # selected options.
+        parsed_value = json.loads(value)
+        parsed_value = (
+            parsed_value if isinstance(parsed_value, list) else [parsed_value]
+        )
+
         selected_choices = {
-            str(v) for v in value if not self._choice_has_empty_value((str(v), ""))
+            stable_json(v)
+            for v in parsed_value
+            if not self._choice_has_empty_value((stable_json(v), ""))
         }
 
         if not self.is_required and not self.allow_multiple_selected:
             default[1].append(self.create_option(name, "", "", False, 0))
 
         choices = [
-            (str(v), json.loads(v)["text"])
-            for v in value
+            (stable_json(v), v["text"])
+            for v in parsed_value
             if v not in (None, "", "null")
         ]
 
         for option_value, option_label in choices:
-            selected = str(option_value) in value and (
+            selected = str(option_value) in parsed_value and (
                 has_selected is False or self.allow_multiple_selected
             )
             has_selected |= selected
@@ -136,7 +142,7 @@ class AutocompleteSelect(Select):
 
     def value_from_datadict(
         self, data: Dict[str, Any], files: Mapping[str, Iterable[Any]], name: str
-    ) -> List[str]:
+    ) -> str:
         """Extract the widget's value from the given data.
 
         Handles the case where multiple values are selected.
@@ -147,51 +153,60 @@ class AutocompleteSelect(Select):
             name: The name of the widget.
 
         Returns:
-            List[str]: A list of JSON string values produced by the widget.
+            str: The JSON value produced by the widget.
         """
         value = super().value_from_datadict(data, files, name)
 
         if value is None:
-            return []
+            value = []
 
-        if not self.allow_multiple_selected:
+        if not isinstance(value, list):
             value = [value]
 
-        parsed_value: List[str] = []
+        parsed_value: List[dict] = []
         for v in value:
             # Skip empty values.
             if v in self.EMPTY_VALUES:
                 continue
 
-            # If the value looks like a JSON object, parse it.
-            if v.startswith(("{", "[")) and v.endswith(("]", "}")):
-                parsed_v = json.loads(v)
-            else:
-                parsed_v = json.loads(stable_json({"id": v, "text": v}))
+            parsed_v = None
 
-            parsed_value.append(parsed_v)
+            # If the value is already a dict, we'll just consider it to be
+            # already parsed.
+            if isinstance(v, dict):
+                parsed_v = v
 
-        return parsed_value
+            # If the value is a string and it looks like a JSON object, parse
+            # it as one.
+            elif isinstance(v, str):
+                if v.startswith("{") and v.endswith("}"):
+                    parsed_v = json.loads(v)
+                else:
+                    parsed_v = json.loads(stable_json({"id": v, "text": v, "value": v}))
+
+            if parsed_v is not None:
+                parsed_value.append(parsed_v)
+
+        # If only one value is expected, unwrap it from the parsed list.
+        final_value = (
+            next(iter(cast(Iterable, parsed_value)), None)
+            if not self.allow_multiple_selected
+            else parsed_value
+        )
+
+        return json.dumps(final_value)
 
     def format_value(self, value: Any) -> Any:
-        """Format a value for rendering.
-
-        Assumes the value is a list of JSON strings. If parsing fails,
-        returns the value as-is.
+        """Format a value for rendering in a widget template.
 
         Args:
             value: The value to format.
 
         Returns:
-            Any: The formatted value, if formatting was successful.
+            Any: The formatted value, if formatting was successful. Otherwise
+                returns the value unmodified.
         """
-        if value in self.EMPTY_VALUES:
-            return []
-
-        try:
-            return [stable_json(v) for v in json.loads(value)]
-        except (json.JSONDecodeError, TypeError):
-            return value
+        return str([] if value in self.EMPTY_VALUES else value)
 
     @staticmethod
     def _choice_has_empty_value(choice: Tuple[Any, str]) -> bool:
